/**************************************************************
 This file is part of
    Morphological Curvature Flow Approximation for AviSynth
                       by Wolfgang Boiger
 Copyright (C) 2007 Wolfgang Boiger, Berlin
 
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 version 2, as published by the Free Software Foundation in
 June 1991 (see 'GNUGPLv2.txt').
 
 There is no guaranty that the rights of third parties
 (including software patents) are not infringed upon by the
 execution or distribution of this program.
 Referring to section 8 of the General Public License, any use
 of this program (including all activities mentioned in
 section 0 and including execution) is restricted to countries
 where any of these activities do not infringe upon any such
 rights of third parties. It lies with the user to verify the
 compliance of his or her use (especially concerning but not
 limited to distribution) of this program with the rights of
 any third parties in the respective countries or regions.
 
 This program is distributed in the hope that it will be
 useful, but WITHOUT ANY WARRANTY; without even the implied
 warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 PURPOSE.  See the GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public
 License 2 along with this program; if not, write to the
   Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 **************************************************************/



#define  VersionString  "2007.03.06"
#define  LegalInfoString  "* Morphological Curvature Flow Approximation for AviSynth by Wolfgang Boiger *\nVersion " VersionString " ;  built  " __DATE__ "  " __TIME__ " .\nCopyright (C) 2007 Wolfgang Boiger, Berlin.\nLicensed under GNU GPL version 2 with possible geographic restrictions.\nThis software comes without any warranty, even without warranty\nof merchantability or fitness for a particular purpose.\nSee files 'AVSCurveFlow.htm' and 'GNUGPLv2.txt' for details."



#include  <stdlib.h>
#include  <math.h>
#include  <string.h>
#include  <ctype.h>

#include  "AviSynth_C.h"



#define  TableIndex_MCF  (0)
#define  TableIndex_MED  (-1)
#define  TableIndex_ACF  (3)
#define  PermutationCount3x3  (2*3*4*5*6*7*8*9)



typedef  struct
{
  unsigned char * TablesEroVector[10];
  unsigned char * TablesDilVector[10];
  unsigned char * * TablesEro;
  unsigned char * * TablesDil;
}
LookupTablesStruct;


typedef  struct
{
  int  Steps;
  unsigned char * Table1;  /* The filter applied first */
  unsigned char * Table2;  /* The filter applied second (or zero) */
  int  Mean;               /* Take the mean value of the results of the first and second filter? Or (if zero) apply them one after the other */
}
CurveFlowDataStruct;



/*****  Function declaration  *****/

typedef  BYTE (* MorphoFunctionType)(BYTE Values[], void * Data);


/** AviSynth stuff **/
void  CurveFlow_Shutdown(void *  Data,  AVS_ScriptEnvironment *  Env);
AVS_Value  AVSC_CC  CurveFlow_Create(AVS_ScriptEnvironment *  Env,  AVS_Value  Args,  void *  Data);
AVS_Value  AVSC_CC  CurveFlowMCF_Create(AVS_ScriptEnvironment *  Env,  AVS_Value  Args,  void *  Data);
AVS_Value  AVSC_CC  CurveFlowACF_Create(AVS_ScriptEnvironment *  Env,  AVS_Value  Args,  void *  Data);
AVS_VideoFrame *  AVSC_CC  CurveFlow_GetFrame(AVS_FilterInfo *  FilterInfo,  int  FrameNo);
void  AVSC_CC  CurveFlow_Free(AVS_FilterInfo *  FilterInfo);

/** Own functions **/
long int  InsertSort(BYTE  Values[],  long int  Index,  int  Start,  int  End);  /* Sorts Values and returns a permutation index */
unsigned char * GetTable(LookupTablesStruct * Tables,  int  TableNumber,  int  Erosion);  /* Will return a pointer to the table requested. Will create the table if necessary */
signed char * JohnsonTrotter9(signed char  Permutation[]);  /* Generates all permutations of [1..9] step by step */

/** Morphological erosion-type functions **/
BYTE  Median(BYTE  Values[], void * area2);
BYTE  ErosionMCF(BYTE  Values[], void * area2);
BYTE  ErosionACF(BYTE  Values[], void * area2);

/** Colorspace dependent functions implemented in AVSShockCSdep.c **/
void  PickPixelYUY2    (unsigned char  PickData[],  const  BYTE  SrcLuma[],  int  Width,  int  Height,  int  SrcPitch,  const  unsigned char  Table1[]);
void  PickPixelYUY2Mean(unsigned char  PickData[],  const  BYTE  SrcLuma[],  int  Width,  int  Height,  int  SrcPitch,  const  unsigned char  Table1[],  const  unsigned char  Table2[]);
void  PickPixelYV12    (unsigned char  PickData[],  const  BYTE  SrcLuma[],  int  Width,  int  Height,  int  SrcPitch,  const  unsigned char  Table1[]);
void  PickPixelYV12Mean(unsigned char  PickData[],  const  BYTE  SrcLuma[],  int  Width,  int  Height,  int  SrcPitch,  const  unsigned char  Table1[],  const  unsigned char  Table2[]);
void  PickPixelRGB     (unsigned char  PickData[],  const  BYTE  SrcData[],  int  Width,  int  Height,  int  SrcPitch,  const  unsigned char  Table1[]);
void  PickPixelRGBMean (unsigned char  PickData[],  const  BYTE  SrcData[],  int  Width,  int  Height,  int  SrcPitch,  const  unsigned char  Table1[],  const  unsigned char  Table2[]);
void  PickPixelRGBA    (unsigned char  PickData[],  const  BYTE  SrcData[],  int  Width,  int  Height,  int  SrcPitch,  const  unsigned char  Table1[]);
void  PickPixelRGBAMean(unsigned char  PickData[],  const  BYTE  SrcData[],  int  Width,  int  Height,  int  SrcPitch,  const  unsigned char  Table1[],  const  unsigned char  Table2[]);
void  ChangeLumaYUY2       (BYTE  DestData[],  const  BYTE  SrcData[],  const  unsigned char  PickData[],  int  Width,  int  Height,  int  DestPitch,  int  SrcPitch);
void  ChangeChromaYUY2     (BYTE  DestData[],  const  BYTE  SrcData[],  const  unsigned char  PickData[],  int  Width,  int  Height,  int  DestPitch,  int  SrcPitch);
void  ChangeLumaYUY2Mean   (BYTE  DestData[],  const  BYTE  SrcData[],  const  unsigned char  PickData[],  int  Width,  int  Height,  int  DestPitch,  int  SrcPitch);
void  ChangeChromaYUY2Mean (BYTE  DestData[],  const  BYTE  SrcData[],  const  unsigned char  PickData[],  int  Width,  int  Height,  int  DestPitch,  int  SrcPitch);
void  ChangeLumaYV12       (BYTE  DestData[],  const  BYTE  SrcData[],  const  unsigned char  PickData[],  int  Width,  int  Height,  int  DestPitch,  int  SrcPitch);
void  ChangeChromaYV12     (BYTE  DestData[],  const  BYTE  SrcData[],  const  unsigned char  PickData[],  int  Width,  int  Height,  int  DestPitch,  int  SrcPitch);
void  ChangeLumaYV12Mean   (BYTE  DestData[],  const  BYTE  SrcData[],  const  unsigned char  PickData[],  int  Width,  int  Height,  int  DestPitch,  int  SrcPitch);
void  ChangeChromaYV12Mean (BYTE  DestData[],  const  BYTE  SrcData[],  const  unsigned char  PickData[],  int  Width,  int  Height,  int  DestPitch,  int  SrcPitch);
void  ChangeChannelRGB     (BYTE  DestData[],  const  BYTE  SrcData[],  const  unsigned char  PickData[],  int  Width,  int  Height,  int  DestPitch,  int  SrcPitch);
void  ChangeChannelRGBMean (BYTE  DestData[],  const  BYTE  SrcData[],  const  unsigned char  PickData[],  int  Width,  int  Height,  int  DestPitch,  int  SrcPitch);
void  ChangeChannelRGBA    (BYTE  DestData[],  const  BYTE  SrcData[],  const  unsigned char  PickData[],  int  Width,  int  Height,  int  DestPitch,  int  SrcPitch);
void  ChangeChannelRGBAMean(BYTE  DestData[],  const  BYTE  SrcData[],  const  unsigned char  PickData[],  int  Width,  int  Height,  int  DestPitch,  int  SrcPitch);



/*****  Function implementation  *****/


AVSC_EXPORT  const char *  AVSC_CC  avisynth_c_plugin_init(AVS_ScriptEnvironment *  Env)
{
  LookupTablesStruct * Tables;
  unsigned short  k;


  /* Initialize (empty) look-up tables */
  Tables = (LookupTablesStruct*)malloc(sizeof(LookupTablesStruct));
  for (k=0 ; k<10 ; k++)  Tables->TablesEroVector[k] = Tables->TablesDilVector[k] = 0;
  Tables->TablesEro = Tables->TablesEroVector+1;
  Tables->TablesDil = Tables->TablesDilVector+1;
  avs_at_exit(Env,CurveFlow_Shutdown,Tables);  /* Shutdown function will free all tables created during filter usage */
  
  avs_add_function(Env,"CurvatureFlow","c.s[Steps]i",CurveFlow_Create,Tables);
  avs_add_function(Env,"MeanCurvatureFlow","c[Steps]i",CurveFlowMCF_Create,Tables);
  avs_add_function(Env,"AffineCurvatureFlow","c[Steps]i",CurveFlowACF_Create,Tables);
  return  "Morphological Curvature Flow Approximation Filter";
}



void  CurveFlow_Shutdown(void *  Data,  AVS_ScriptEnvironment *  Env)
{
  LookupTablesStruct * Tables;
  unsigned short  k;


  if (Data)
  {
    Tables = (LookupTablesStruct*)Data;
    for (k=0 ; k<10 ; k++)  if (Tables->TablesEroVector[k])  free(Tables->TablesEroVector[k]);
    for (k=0 ; k<10 ; k++)  if (Tables->TablesDilVector[k])  free(Tables->TablesDilVector[k]);
    free(Tables);
  }
}



AVS_Value  AVSC_CC  CurveFlow_Create(AVS_ScriptEnvironment *  Env,  AVS_Value  Args,  void *  Data)
{
  AVS_FilterInfo * FilterInfo;
  AVS_VideoInfo * VideoInfo;
  AVS_Value  SrcClipValue, DestClipValue, TypeValue, ModeValue, StepsValue;
  AVS_Clip * DestClip;
  CurveFlowDataStruct * CurveFlowData;
  char * ErrorText;
  char  TempString[10];  /* Long enough to take all possible text arguments */
  int  Steps, Type;
  char  Mode;
  int  k;


  SrcClipValue = avs_array_elt(Args,0);
  TypeValue = avs_array_elt(Args,1);
  ModeValue = avs_array_elt(Args,2);
  StepsValue = avs_array_elt(Args,3);
  
  /* Check arguments */
  /* Steps */
  Steps = (avs_is_int(StepsValue)?avs_as_int(StepsValue):1);  /* (Steps default value) */
  if (Steps<0)  return  avs_new_value_error("CurvatureFlow: Number of Steps may not be negative");
  /* Type */
  Type = -2;
  if (avs_is_string(TypeValue))
  {
    strncpy(TempString,avs_as_string(TypeValue),9);
    TempString[9] = 0;
    for (k=0 ; k<10 ; k++)  TempString[k] = toupper(TempString[k]);
    if (!strcmp(TempString,"I"))  Type = 1;
    if (!strcmp(TempString,"II"))  Type = 2;
    if (!strcmp(TempString,"III"))  Type = 3;
    if (!strcmp(TempString,"IV"))  Type = 4;
    if (!strcmp(TempString,"V"))  Type = 5;
    if (!strcmp(TempString,"VI"))  Type = 6;
    if (!strcmp(TempString,"VII"))  Type = 7;
    if (!strcmp(TempString,"VIII"))  Type = 8;
    if (!strcmp(TempString,"ACF"))  Type = TableIndex_ACF;
    if (!strcmp(TempString,"AMSS"))  Type = TableIndex_ACF;
    if (!strcmp(TempString,"MCF"))  Type = TableIndex_MCF;
    if (!strcmp(TempString,"MCM"))  Type = TableIndex_MCF;
    if (!strcmp(TempString,"MED"))  Type = TableIndex_MED;
    if (!strcmp(TempString,"MEDIAN"))  Type = TableIndex_MED;
    if (!strcmp(TempString,"INFO"))  return  avs_new_value_error(LegalInfoString);
    if (Type==-2)  return  avs_new_value_error("CurvatureFlow: Type unknown");
  }
  if (avs_is_int(TypeValue))  Type = avs_as_int(TypeValue);
  if (!(avs_is_int(TypeValue) || avs_is_string(TypeValue)))  return  avs_new_value_error("CurvatureFlow: Type argument may only be a integer or a string");
  if (Type<-1 || Type>9)  return  avs_new_value_error("CurvatureFlow: Type out of range (only -1..9 allowed)");
  if (Type==9)  Type = 8;
  /* Mode */
  Mode = 'D';
  if (Type!=TableIndex_MED)
  {
    Mode = 0;
    strncpy(TempString,avs_as_string(ModeValue),9);
    TempString[9] = 0;
    for (k=0 ; k<10 ; k++)  TempString[k] = toupper(TempString[k]);
    if (!strcmp(TempString,"DE"))  Mode = 'O';
    if (!strcmp(TempString,"O"))  Mode = 'O';
    if (!strcmp(TempString,"ED"))  Mode = 'C';
    if (!strcmp(TempString,"C"))  Mode = 'C';
    if (!strcmp(TempString,"D"))  Mode = 'D';
    if (!strcmp(TempString,"E"))  Mode = 'E';
    if (!strcmp(TempString,"M"))  Mode = 'M';
    if (!strcmp(TempString,"MEAN"))  Mode = 'M';
    if (!Mode)  return  avs_new_value_error("CurvatureFlow: Mode unknown");
  }
  
  /* Create filter, check properties of input clip */
  DestClip = avs_new_c_filter(Env,&FilterInfo,SrcClipValue,1);
  if (!DestClip)  return  avs_new_value_error("CurvatureFlow: Could not create filter");
  VideoInfo = &FilterInfo->vi;
  /* If clip is only one pixel or no video at all, if Steps vanishes, or if Type==1 (!!) , we return the clip unchanged (no error in this case) */
  if (!avs_has_video(VideoInfo) || (VideoInfo->width==0) || (VideoInfo->height==0) || ((VideoInfo->width==1) && (VideoInfo->height==1)) || (Steps==0) || (Type==1))
  {
    DestClipValue = avs_new_value_clip(DestClip);
    avs_release_clip(DestClip);
    if (!avs_defined(DestClipValue))  return  avs_new_value_error("CurvatureFlow: Could not create new video clip");
    return  DestClipValue;
  }

  CurveFlowData = 0;
  ErrorText = 0;
  /* From now on, when an error is detected, a message will be stored in ErrorText. Hence (ErrorText) indicates that something is wrong. */
  if (!(avs_is_yuy2(VideoInfo) || avs_is_yv12(VideoInfo) || avs_is_rgb24(VideoInfo) || avs_is_rgb32(VideoInfo)))  ErrorText = "CurvatureFlow: Only YUY2, YV12, or RGB24, RGB32 color spaces allowed";  /* If new AviSynth versions support new color spaces, we don't support them (yet) */
  if (!ErrorText)  CurveFlowData = (CurveFlowDataStruct*)malloc(sizeof(CurveFlowDataStruct));  /* Allocate my data structure */
  if (!ErrorText && !CurveFlowData)  ErrorText = "CurvatureFlow: Could not allocate memory";
  if (!ErrorText)
  {
    CurveFlowData->Steps = Steps;
    CurveFlowData->Mean = (Mode=='M');
    CurveFlowData->Table1 = 0;
    CurveFlowData->Table2 = 0;
    /* Guided by the Mode, this decides which lookup table is used in which order */
    if (Mode=='D' || Mode=='C' || Mode=='M')  CurveFlowData->Table1 = GetTable((LookupTablesStruct*)Data,Type,0);
    if (Mode=='E' || Mode=='O')  CurveFlowData->Table1 = GetTable((LookupTablesStruct*)Data,Type,1);
    if (Mode=='C' || Mode=='M')  CurveFlowData->Table2 = GetTable((LookupTablesStruct*)Data,Type,1);
    if (Mode=='O')  CurveFlowData->Table2 = GetTable((LookupTablesStruct*)Data,Type,0);
    if (!CurveFlowData->Table1 || ((Mode=='M' || Mode=='O' || Mode=='C') && !CurveFlowData->Table1))  ErrorText = "CurvatureFlow: Look-up table generation failed (out of memory?)";
  }
  if (!ErrorText)
  {
    FilterInfo->user_data = CurveFlowData;  /* Structure will be available to (and needed by) GetFrame function */
    FilterInfo->get_frame = CurveFlow_GetFrame;  /* Change frame-computation-function of new clip */
    FilterInfo->free_filter = CurveFlow_Free;  /* Will be called, when filter is destroyed, frees memory occupied by ShockData */
    DestClipValue = avs_new_value_clip(DestClip);  /* Create return value for this function */
    if (!avs_defined(DestClipValue))  ErrorText = "CurvatureFlow: Could not create new video clip";
  }
  /* No error up to this point => We're done! */
  avs_release_clip(DestClip);
  if (!ErrorText)  return  DestClipValue;
  /* Error occured: Clean up memory, return error message */
  // TODO Release return clip value
  if (CurveFlowData)  free(CurveFlowData);
  return  avs_new_value_error(ErrorText);
}



AVS_Value  AVSC_CC  CurveFlowMCF_Create(AVS_ScriptEnvironment *  Env,  AVS_Value  Args,  void *  Data)
{
  AVS_Value  FullArgs[5];


  /* We prepare arguments for the general CurvatureFlow function, so it acts like the MCF-Type */
  FullArgs[0] = avs_array_elt(Args,0);
  FullArgs[1] = avs_new_value_string("MCF");
  FullArgs[2] = avs_new_value_string("M");
  FullArgs[3] = avs_array_elt(Args,1);
  FullArgs[4] = avs_array_elt(Args,2);
  return  CurveFlow_Create(Env,avs_new_value_array(FullArgs,5),Data);
}



AVS_Value  AVSC_CC  CurveFlowACF_Create(AVS_ScriptEnvironment *  Env,  AVS_Value  Args,  void *  Data)
{
  AVS_Value  FullArgs[5];


  /* We prepare arguments for the general CurvatureFlow function, so it acts like the ACF-Type */
  FullArgs[0] = avs_array_elt(Args,0);
  FullArgs[1] = avs_new_value_string("ACF");
  FullArgs[2] = avs_new_value_string("M");
  FullArgs[3] = avs_array_elt(Args,1);
  FullArgs[4] = avs_array_elt(Args,2);
  return  CurveFlow_Create(Env,avs_new_value_array(FullArgs,5),Data);
}



AVS_VideoFrame *  AVSC_CC  CurveFlow_GetFrame(AVS_FilterInfo *  FilterInfo,  int  FrameNo)
{
  /* Filter data */
  int  Steps, Mean;
  unsigned char * Tables[2];
  /* Video frames */
  int  Width, Height, PixelType;
  AVS_VideoFrame * Frame[2];
  BYTE * DataY[2], * DataU[2], * DataV[2];
  #define  Data  DataY
  int  PitchY[2], PitchU[2], PitchV[2];
  #define  Pitch  PitchY
  /* Temporary memory */
  unsigned char * PickData;
  char * ErrorText;
  int  k;


  Steps = ((CurveFlowDataStruct*)FilterInfo->user_data)->Steps;
  Mean = ((CurveFlowDataStruct*)FilterInfo->user_data)->Mean;
  Tables[0] = ((CurveFlowDataStruct*)FilterInfo->user_data)->Table1;
  Tables[1] = ((CurveFlowDataStruct*)FilterInfo->user_data)->Table2;
  if (Tables[1] && !Mean)  Steps *= 2;
  if (!Tables[1])  Mean = 0;
  if (!Tables[1])  Tables[1] = Tables[0];
  
  Width = FilterInfo->vi.width;
  Height = FilterInfo->vi.height;
  PixelType = FilterInfo->vi.pixel_type;
  
  ErrorText = 0;
  Frame[1] = 0;
  PickData = 0;
  Frame[0] = avs_get_frame(FilterInfo->child,FrameNo);
  if (!Frame[0])  ErrorText = "CurvatureFlow: Could not get source frame";
  if (!ErrorText)
  {
    /* PickData will store which Tabel picked which neighbor for which pixel */
    PickData = (unsigned char*)malloc(sizeof(unsigned char)*Width*Height);
    if (!PickData)  ErrorText = "CurvatureFlow: Could not allocate temporary memory";
  }
  /* We make the source frame writable and allocate another frame - then each iteration transfers data between those frames */
  if (!ErrorText)
  {
    avs_make_writable(FilterInfo->env,&(Frame[0]));
    if (!Frame[0])  ErrorText = "CurvatureFlow: Could make source frame writable";
  }
  if (!ErrorText)
  {
    Frame[1] = avs_new_video_frame(FilterInfo->env,&FilterInfo->vi);
    if (!Frame[0])  ErrorText = "CurvatureFlow: Could not get second frame";
  }
  if (!ErrorText)
  {
    /* Depending on the color space, we gather the data pointers */
    switch (FilterInfo->vi.pixel_type)
    {
      case AVS_CS_YV12:
      case AVS_CS_I420:
        for (k=0 ; k<2 ; k++)
        {
          DataY[k] = avs_get_write_ptr_p(Frame[k],AVS_PLANAR_Y);
          DataU[k] = avs_get_write_ptr_p(Frame[k],AVS_PLANAR_U);
          DataV[k] = avs_get_write_ptr_p(Frame[k],AVS_PLANAR_V);
          PitchY[k] = avs_get_pitch_p(Frame[k],AVS_PLANAR_Y);
          PitchU[k] = avs_get_pitch_p(Frame[k],AVS_PLANAR_U);
          PitchV[k] = avs_get_pitch_p(Frame[k],AVS_PLANAR_V);
          if (!DataY[k] || !DataU[k] || !DataV[k])  ErrorText = "CurvatureFlow: Could not get data pointers";
        }
      break;
      case AVS_CS_YUY2:
      case AVS_CS_BGR24:
      case AVS_CS_BGR32:
        for (k=0 ; k<2 ; k++)
        {
          Data[k] = avs_get_write_ptr(Frame[k]);
          Pitch[k] = avs_get_pitch(Frame[k]);
          if (!Data[k])  ErrorText = "CurvatureFlow: Could not get data pointers";
        }
      break;
      default:
        /*Data[0] = */ DataY[0] = DataU[0] = DataV[0] = 0;
        /*Data[1] = */ DataY[1] = DataU[1] = DataV[1] = 0;
        ErrorText = "CurvatureFlow internal error: Color space unknown / not allowed";  /* Uhou ?? */
    }
  }
  if (!ErrorText)
  {
    /** Here the core functions are called **/
    /* Source frame = Frame[k%2]    target frame = Frame[(k+1)%2] */
    if (Mean)
    {
      switch (FilterInfo->vi.pixel_type)
      {
        case AVS_CS_YV12:
        case AVS_CS_I420:
          for (k=0 ; k<Steps ; k++)
          {
            PickPixelYV12Mean(PickData,DataY[k%2],Width,Height,PitchY[k%2],Tables[0],Tables[1]);
            ChangeLumaYV12Mean(DataY[(k+1)%2],DataY[k%2],PickData,Width,Height,PitchY[(k+1)%2],PitchY[k%2]);
            ChangeChromaYV12Mean(DataU[(k+1)%2],DataU[k%2],PickData,Width,Height,PitchU[(k+1)%2],PitchU[k%2]);
            ChangeChromaYV12Mean(DataV[(k+1)%2],DataV[k%2],PickData,Width,Height,PitchV[(k+1)%2],PitchV[k%2]);
          }
        break;
        case AVS_CS_YUY2:
          for (k=0 ; k<Steps ; k++)
          {
            PickPixelYUY2Mean(PickData,Data[k%2],Width,Height,Pitch[k%2],Tables[0],Tables[1]);
            ChangeLumaYUY2Mean(Data[(k+1)%2],Data[k%2],PickData,Width,Height,Pitch[(k+1)%2],Pitch[k%2]);
            ChangeChromaYUY2Mean(Data[(k+1)%2]+1,Data[k%2]+1,PickData,Width,Height,Pitch[(k+1)%2],Pitch[k%2]);
            ChangeChromaYUY2Mean(Data[(k+1)%2]+3,Data[k%2]+3,PickData,Width,Height,Pitch[(k+1)%2],Pitch[k%2]);
          }
        break;
        case AVS_CS_BGR24:
          for (k=0 ; k<Steps ; k++)
          {
            PickPixelRGBMean(PickData,Data[k%2],Width,Height,Pitch[k%2],Tables[0],Tables[1]);
            ChangeChannelRGBMean(Data[(k+1)%2],Data[k%2],PickData,Width,Height,Pitch[(k+1)%2],Pitch[k%2]);
            PickPixelRGBMean(PickData,Data[k%2]+1,Width,Height,Pitch[k%2],Tables[0],Tables[1]);
            ChangeChannelRGBMean(Data[(k+1)%2]+1,Data[k%2]+1,PickData,Width,Height,Pitch[(k+1)%2],Pitch[k%2]);
            PickPixelRGBMean(PickData,Data[k%2]+2,Width,Height,Pitch[k%2],Tables[0],Tables[1]);
            ChangeChannelRGBMean(Data[(k+1)%2]+2,Data[k%2]+2,PickData,Width,Height,Pitch[(k+1)%2],Pitch[k%2]);
          }
        break;
        case AVS_CS_BGR32:
          for (k=0 ; k<Steps ; k++)
          {
            PickPixelRGBAMean(PickData,Data[k%2],Width,Height,Pitch[k%2],Tables[0],Tables[1]);
            ChangeChannelRGBAMean(Data[(k+1)%2],Data[k%2],PickData,Width,Height,Pitch[(k+1)%2],Pitch[k%2]);
            PickPixelRGBAMean(PickData,Data[k%2]+1,Width,Height,Pitch[k%2],Tables[0],Tables[1]);
            ChangeChannelRGBAMean(Data[(k+1)%2]+1,Data[k%2]+1,PickData,Width,Height,Pitch[(k+1)%2],Pitch[k%2]);
            PickPixelRGBAMean(PickData,Data[k%2]+2,Width,Height,Pitch[k%2],Tables[0],Tables[1]);
            ChangeChannelRGBAMean(Data[(k+1)%2]+2,Data[k%2]+2,PickData,Width,Height,Pitch[(k+1)%2],Pitch[k%2]);
            PickPixelRGBAMean(PickData,Data[k%2]+3,Width,Height,Pitch[k%2],Tables[0],Tables[1]);
            ChangeChannelRGBAMean(Data[(k+1)%2]+3,Data[k%2]+3,PickData,Width,Height,Pitch[(k+1)%2],Pitch[k%2]);
          }
        break;
      }
    }
    else
    {
      switch (FilterInfo->vi.pixel_type)
      {
        case AVS_CS_YV12:
        case AVS_CS_I420:
          for (k=0 ; k<Steps ; k++)
          {
            PickPixelYV12(PickData,DataY[k%2],Width,Height,PitchY[k%2],Tables[k%2]);
            ChangeLumaYV12(DataY[(k+1)%2],DataY[k%2],PickData,Width,Height,PitchY[(k+1)%2],PitchY[k%2]);
            ChangeChromaYV12(DataU[(k+1)%2],DataU[k%2],PickData,Width,Height,PitchU[(k+1)%2],PitchU[k%2]);
            ChangeChromaYV12(DataV[(k+1)%2],DataV[k%2],PickData,Width,Height,PitchV[(k+1)%2],PitchV[k%2]);
          }
        break;
        case AVS_CS_YUY2:
          for (k=0 ; k<Steps ; k++)
          {
            PickPixelYUY2(PickData,Data[k%2],Width,Height,Pitch[k%2],Tables[k%2]);
            ChangeLumaYUY2(Data[(k+1)%2],Data[k%2],PickData,Width,Height,Pitch[(k+1)%2],Pitch[k%2]);
            ChangeChromaYUY2(Data[(k+1)%2]+1,Data[k%2]+1,PickData,Width,Height,Pitch[(k+1)%2],Pitch[k%2]);
            ChangeChromaYUY2(Data[(k+1)%2]+3,Data[k%2]+3,PickData,Width,Height,Pitch[(k+1)%2],Pitch[k%2]);
          }
        break;
        case AVS_CS_BGR24:
          for (k=0 ; k<Steps ; k++)
          {
            PickPixelRGB(PickData,Data[k%2],Width,Height,Pitch[k%2],Tables[k%2]);
            ChangeChannelRGB(Data[(k+1)%2],Data[k%2],PickData,Width,Height,Pitch[(k+1)%2],Pitch[k%2]);
            PickPixelRGB(PickData,Data[k%2]+1,Width,Height,Pitch[k%2],Tables[k%2]);
            ChangeChannelRGB(Data[(k+1)%2]+1,Data[k%2]+1,PickData,Width,Height,Pitch[(k+1)%2],Pitch[k%2]);
            PickPixelRGB(PickData,Data[k%2]+2,Width,Height,Pitch[k%2],Tables[k%2]);
            ChangeChannelRGB(Data[(k+1)%2]+2,Data[k%2]+2,PickData,Width,Height,Pitch[(k+1)%2],Pitch[k%2]);
          }
        break;
        case AVS_CS_BGR32:
          for (k=0 ; k<Steps ; k++)
          {
            PickPixelRGBA(PickData,Data[k%2],Width,Height,Pitch[k%2],Tables[k%2]);
            ChangeChannelRGBA(Data[(k+1)%2],Data[k%2],PickData,Width,Height,Pitch[(k+1)%2],Pitch[k%2]);
            PickPixelRGBA(PickData,Data[k%2]+1,Width,Height,Pitch[k%2],Tables[k%2]);
            ChangeChannelRGBA(Data[(k+1)%2]+1,Data[k%2]+1,PickData,Width,Height,Pitch[(k+1)%2],Pitch[k%2]);
            PickPixelRGBA(PickData,Data[k%2]+2,Width,Height,Pitch[k%2],Tables[k%2]);
            ChangeChannelRGBA(Data[(k+1)%2]+2,Data[k%2]+2,PickData,Width,Height,Pitch[(k+1)%2],Pitch[k%2]);
            PickPixelRGBA(PickData,Data[k%2]+3,Width,Height,Pitch[k%2],Tables[k%2]);
            ChangeChannelRGBA(Data[(k+1)%2]+3,Data[k%2]+3,PickData,Width,Height,Pitch[(k+1)%2],Pitch[k%2]);
          }
        break;
      }
    }
  }
  /* Free everything (whether error or not) */
  if (PickData)  free(PickData);
  if (Frame[(Steps+1)%2])  avs_release_video_frame(Frame[(Steps+1)%2]);
  if (ErrorText)
  {
    /* If error the final target frame is also released */
    if (Frame[Steps%2])  avs_release_video_frame(Frame[Steps%2]);
    Frame[Steps%2] = 0;
    FilterInfo->error = ErrorText;
  }
  return  Frame[Steps%2];

  #undef  Data
  #undef  Pitch
}



void  AVSC_CC  CurveFlow_Free(AVS_FilterInfo *  FilterInfo)
{
  if (FilterInfo->user_data)  free(FilterInfo->user_data);
  FilterInfo->user_data = 0;
}



long int  InsertSort(BYTE  Values[],  long int  Index,  int  Start,  int  End)
/* Assumes Values[0..Start-1] to be sorted (Start >= 1) with permutation index Index.
   Sorts until Values[0..End-1] is sorted and return new permutation index            */
{
  int  k, j;
  BYTE  Value;


  for (k=Start ; k<End ; k++)
  {
    Index *= k+1;
    for (j=k ; j&&(Values[j-1]>Values[k]) ; j--) ;
    if (j<k)
    {
      Value = Values[k];
      memmove(Values+j+1,Values+j,sizeof(BYTE)*(k-j));
      Values[j] = Value;
      Index += k-j;
    }
  }
  return  Index;
}



unsigned char * GetTable(LookupTablesStruct * Tables,  int  TableNumber,  int  Erosion)
{
  unsigned char * * TablePointer;
  unsigned char * Table;
  MorphoFunctionType  MorphoFcn;
  signed char  Permutation[9];
  long int  Index;
  BYTE  Values[9];
  BYTE  Value;
  int  k;


  TablePointer = Erosion?Tables->TablesEro:Tables->TablesDil;
  TablePointer += TableNumber;
  if (TablePointer[0])  return  TablePointer[0];
  /* OK, the table hasn't been built before */
  TablePointer[0] = (unsigned char*)malloc(sizeof(unsigned char)*PermutationCount3x3);
  Table = TablePointer[0];
  if (!Table)  return  0;
  /* and memory has been allocated successfully, so we build it... */
  
  /* ** Here we decide which erosion function is used to build the table, based on the TableNumber ** */
  MorphoFcn = ErosionACF;
  if (TableNumber==TableIndex_MCF)  MorphoFcn = ErosionMCF;
  if (TableNumber==TableIndex_MED)  MorphoFcn = Median;
  
  Erosion = Erosion?(+1):(-1);
  Permutation[0] = 0;
  while (JohnsonTrotter9(Permutation))
  {
    for (k=0 ; k<9 ; k++)  Values[k] = abs(Permutation[k]);
    Index = InsertSort(Values,0,1,9);
    /* Right pixel / least significant bits */
    for (k=0 ; k<9 ; k++)  Values[k] = 10+Erosion*abs(Permutation[k]);
    Value = abs(10-(*MorphoFcn)(Values,&TableNumber));
    for (k=0 ; Value!=abs(Permutation[k]) ; k++) ;
    Table[Index] = (k%3)*4+k/3;
    /* Left pixel / most significant bits */
    /*************************************************************
      Morpho-Value-Order                        Insert-Sort-Order
       0  3  6               == (x+6)%9 ==>              6  0  3
       1  4  7                                           7  1  4
       2  5  8              <== (x+3)%9 ==               8  2  5
     *************************************************************/
    for (k=0 ; k<9 ; k++)  Values[k] = 10+Erosion*abs(Permutation[(k+6)%9]);
    Value = abs(10-(*MorphoFcn)(Values,&TableNumber));
    for (k=0 ; Value!=abs(Permutation[k]) ; k++) ;
    k = (k+3)%9;
    Table[Index] += 16*((k%3)*4+k/3);
  }
  return  Table;
}



signed char * JohnsonTrotter9(signed char  Permutation[])
/* Johnson-Trotter-Algorithm
   (http://www.cut-the-knot.org/Curriculum/Combinatorics/JohnsonTrotter.shtml, 2006-01-30)
   Generates all permutations of S_n (here n=9) step by step
   Initialize with Permutation[0] = 0
   Each call will place the next permutation in Permutation
   Values are in the range 1..n
   Signs are for internal use - read values with abs(Permutation[k])
   Last call will return NULL (indicating all permutations have appeared)                   */
{
  #define  Sign(X)  (((X)>0)?1:(((X)==0)?0:(-1)))

  signed char  k, Mobile, Figure;
  
  
  if (!(Permutation[0]))
  {
    for (k=0 ; k<9 ; k++)  Permutation[k] = -k-1;
    return  Permutation;
  }
  Mobile = -1;
  for (k=0 ; k<9 ; k++)
   if ( ((Mobile==-1) || (abs(Permutation[Mobile])<abs(Permutation[k])))     /* There is no mobile yet  OR  the current figure is bigger, than the last biggest mobile number found */
                                                                             /* AND    the current figure is mobile, i.e.: */
        && ((k>0) || (Sign(Permutation[k])==1))                              /*  if it is the first one, it needs to look to the right, */
        && ((k<8) || (Sign(Permutation[k])==-1))                             /*  if it is the last one, it needs to look to the left, */
        && (abs(Permutation[k+Sign(Permutation[k])])<abs(Permutation[k])) )  /*  it needs to be bigger, than the figure it is looking at. */
   Mobile = k;
  if (Mobile==-1)  return  0;
  k = Mobile+Sign(Permutation[Mobile]);
  Figure = Permutation[Mobile];
  Permutation[Mobile] = Permutation[k];
  Permutation[k] = Figure;
  Figure = abs(Figure);
  for (k=0 ; k<9 ; k++)
   if (abs(Permutation[k]) > Figure)  Permutation[k] *= -1;
  return  Permutation;

  #undef  Sign
}



BYTE  Median(BYTE  Values[], void * area2)
{
  InsertSort(Values,0,1,9);
  return  Values[4];
}



BYTE  ErosionMCF(BYTE  Values[], void * area2)
{
  #define  Min(A,B)  ((A)<(B)?(A):(B))
  #define  Max(A,B)  ((A)>(B)?(A):(B))
  return  Max(Max(Min(Min(Values[0],Values[4]),Values[8]),
                  Min(Min(Values[3],Values[4]),Values[5])),
              Max(Min(Min(Values[6],Values[4]),Values[2]),
                  Min(Min(Values[7],Values[4]),Values[1])));
  #undef  Min
  #undef  Max
}



BYTE  ErosionACF(BYTE  Values[], void * area2)
{
  BYTE  Circle[12];
  BYTE  Center, Value;
  int  j, k, Area;


  Area = (*((int*)area2))-1;
  Center = Values[4];
  Circle[0] = Values[0];  /* Circle stores all the 8 neighbor values in a cyclic way, so we can go around with all half-planes */
  Circle[1] = Values[1];
  Circle[2] = Values[2];
  Circle[3] = Values[5];
  Circle[4] = Values[8];
  Circle[5] = Values[7];
  Circle[6] = Values[6];
  Circle[7] = Values[3];
  memcpy(Circle+8,Circle,sizeof(BYTE)*4);
  InsertSort(Values,0,1,9);  /* All possible values to return are stored in Values, so we sort this vector */
  /* This is the Local Affine Erosion */
  for (k=8 ; k ; k--)
  {
    Value = Values[k];
    if (Center < Value)  continue;
    j = 8;
    while (j--)
     if (((Circle[j  ]>=Value)*1 +
          (Circle[j+1]>=Value)*2 +
          (Circle[j+2]>=Value)*2 +
          (Circle[j+3]>=Value)*2 +
          (Circle[j+4]>=Value)*1   ) < Area)  break;
    if (j==-1)  break;
  }
  return  Values[k];
}



/**************************************************************
 This part implements all colorspace dependent functions from
 AVSCurveFlowCSdep.c.
 **************************************************************/

/***** PickPixel *****/
#define  AVSCurveFlow_CSdep_PickPixel  PickPixelYUY2
#define  AVSCurveFlow_CSdep_Stride  2
#define  AVSCurveFlow_CSdep_Table2  0
#define  AVSCurveFlow_CSdep_WidthEven  1
#include  "AVSCurveFlowCSdep.c"

#define  AVSCurveFlow_CSdep_PickPixel  PickPixelYUY2Mean
#define  AVSCurveFlow_CSdep_Stride  2
#define  AVSCurveFlow_CSdep_Table2  1
#define  AVSCurveFlow_CSdep_WidthEven  1
#include  "AVSCurveFlowCSdep.c"

#define  AVSCurveFlow_CSdep_PickPixel  PickPixelYV12
#define  AVSCurveFlow_CSdep_Stride  1
#define  AVSCurveFlow_CSdep_Table2  0
#define  AVSCurveFlow_CSdep_WidthEven  1
#include  "AVSCurveFlowCSdep.c"

#define  AVSCurveFlow_CSdep_PickPixel  PickPixelYV12Mean
#define  AVSCurveFlow_CSdep_Stride  1
#define  AVSCurveFlow_CSdep_Table2  1
#define  AVSCurveFlow_CSdep_WidthEven  1
#include  "AVSCurveFlowCSdep.c"

#define  AVSCurveFlow_CSdep_PickPixel  PickPixelRGB
#define  AVSCurveFlow_CSdep_Stride  3
#define  AVSCurveFlow_CSdep_Table2  0
#define  AVSCurveFlow_CSdep_WidthEven  0
#include  "AVSCurveFlowCSdep.c"

#define  AVSCurveFlow_CSdep_PickPixel  PickPixelRGBMean
#define  AVSCurveFlow_CSdep_Stride  3
#define  AVSCurveFlow_CSdep_Table2  1
#define  AVSCurveFlow_CSdep_WidthEven  0
#include  "AVSCurveFlowCSdep.c"

#define  AVSCurveFlow_CSdep_PickPixel  PickPixelRGBA
#define  AVSCurveFlow_CSdep_Stride  4
#define  AVSCurveFlow_CSdep_Table2  0
#define  AVSCurveFlow_CSdep_WidthEven  0
#include  "AVSCurveFlowCSdep.c"

#define  AVSCurveFlow_CSdep_PickPixel  PickPixelRGBAMean
#define  AVSCurveFlow_CSdep_Stride  4
#define  AVSCurveFlow_CSdep_Table2  1
#define  AVSCurveFlow_CSdep_WidthEven  0
#include  "AVSCurveFlowCSdep.c"

/***** ChangePixel *****/
#define  AVSCurveFlow_CSdep_ChangePixel  ChangeLumaYUY2
#define  AVSCurveFlow_CSdep_Stride  2
#define  AVSCurveFlow_CSdep_ChromaWidth  1
#define  AVSCurveFlow_CSdep_ChromaHeight  1
#define  AVSCurveFlow_CSdep_Table2  0
#include  "AVSCurveFlowCSdep.c"

#define  AVSCurveFlow_CSdep_ChangePixel  ChangeChromaYUY2
#define  AVSCurveFlow_CSdep_Stride  4
#define  AVSCurveFlow_CSdep_ChromaWidth  2
#define  AVSCurveFlow_CSdep_ChromaHeight  1
#define  AVSCurveFlow_CSdep_Table2  0
#include  "AVSCurveFlowCSdep.c"

#define  AVSCurveFlow_CSdep_ChangePixel  ChangeLumaYUY2Mean
#define  AVSCurveFlow_CSdep_Stride  2
#define  AVSCurveFlow_CSdep_ChromaWidth  1
#define  AVSCurveFlow_CSdep_ChromaHeight  1
#define  AVSCurveFlow_CSdep_Table2  1
#include  "AVSCurveFlowCSdep.c"

#define  AVSCurveFlow_CSdep_ChangePixel  ChangeChromaYUY2Mean
#define  AVSCurveFlow_CSdep_Stride  4
#define  AVSCurveFlow_CSdep_ChromaWidth  2
#define  AVSCurveFlow_CSdep_ChromaHeight  1
#define  AVSCurveFlow_CSdep_Table2  1
#include  "AVSCurveFlowCSdep.c"

#define  AVSCurveFlow_CSdep_ChangePixel  ChangeLumaYV12
#define  AVSCurveFlow_CSdep_Stride  1
#define  AVSCurveFlow_CSdep_ChromaWidth  1
#define  AVSCurveFlow_CSdep_ChromaHeight  1
#define  AVSCurveFlow_CSdep_Table2  0
#include  "AVSCurveFlowCSdep.c"

#define  AVSCurveFlow_CSdep_ChangePixel  ChangeChromaYV12
#define  AVSCurveFlow_CSdep_Stride  1
#define  AVSCurveFlow_CSdep_ChromaWidth  2
#define  AVSCurveFlow_CSdep_ChromaHeight  2
#define  AVSCurveFlow_CSdep_Table2  0
#include  "AVSCurveFlowCSdep.c"

#define  AVSCurveFlow_CSdep_ChangePixel  ChangeLumaYV12Mean
#define  AVSCurveFlow_CSdep_Stride  1
#define  AVSCurveFlow_CSdep_ChromaWidth  1
#define  AVSCurveFlow_CSdep_ChromaHeight  1
#define  AVSCurveFlow_CSdep_Table2  1
#include  "AVSCurveFlowCSdep.c"

#define  AVSCurveFlow_CSdep_ChangePixel  ChangeChromaYV12Mean
#define  AVSCurveFlow_CSdep_Stride  1
#define  AVSCurveFlow_CSdep_ChromaWidth  2
#define  AVSCurveFlow_CSdep_ChromaHeight  2
#define  AVSCurveFlow_CSdep_Table2  1
#include  "AVSCurveFlowCSdep.c"

#define  AVSCurveFlow_CSdep_ChangePixel  ChangeChannelRGB
#define  AVSCurveFlow_CSdep_Stride  3
#define  AVSCurveFlow_CSdep_ChromaWidth  1
#define  AVSCurveFlow_CSdep_ChromaHeight  1
#define  AVSCurveFlow_CSdep_Table2  0
#include  "AVSCurveFlowCSdep.c"

#define  AVSCurveFlow_CSdep_ChangePixel  ChangeChannelRGBMean
#define  AVSCurveFlow_CSdep_Stride  3
#define  AVSCurveFlow_CSdep_ChromaWidth  1
#define  AVSCurveFlow_CSdep_ChromaHeight  1
#define  AVSCurveFlow_CSdep_Table2  1
#include  "AVSCurveFlowCSdep.c"

#define  AVSCurveFlow_CSdep_ChangePixel  ChangeChannelRGBA
#define  AVSCurveFlow_CSdep_Stride  4
#define  AVSCurveFlow_CSdep_ChromaWidth  1
#define  AVSCurveFlow_CSdep_ChromaHeight  1
#define  AVSCurveFlow_CSdep_Table2  0
#include  "AVSCurveFlowCSdep.c"

#define  AVSCurveFlow_CSdep_ChangePixel  ChangeChannelRGBAMean
#define  AVSCurveFlow_CSdep_Stride  4
#define  AVSCurveFlow_CSdep_ChromaWidth  1
#define  AVSCurveFlow_CSdep_ChromaHeight  1
#define  AVSCurveFlow_CSdep_Table2  1
#include  "AVSCurveFlowCSdep.c"

/**************************************************************/
